/* 
This work is licensed under the Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License. 
To view a copy of this license, visit http://creativecommons.org/licenses/by-nc-sa/4.0/ or send a letter to Creative Commons, PO Box 1866, Mountain View, CA 94042, USA.
*/

#include <mitsuba/render/bsdf.h>
#include <mitsuba/hw/basicshader.h>
#include <mitsuba/core/warp.h>
#include <mitsuba/core/bitmap.h>
#include <mitsuba/core/fresolver.h>
#include <mitsuba/core/fstream.h>
#include <mitsuba/core/mstream.h>
#include <mitsuba/core/plugin.h>
#include <mitsuba/core/sched.h>
#include <mitsuba/render/texture.h>
#include <mitsuba/render/mipmap.h>
#include <mitsuba/hw/renderer.h>
#include <mitsuba/hw/gputexture.h>
#include <mitsuba/hw/gpuprogram.h>
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>
#include <math.h> 
#include <iostream>
#include <string>
#include <iterator>
// Matrix multiplication library
#include <libxsmm_source.h>  
MTS_NAMESPACE_BEGIN 

inline BSDFSamplingRecord::BSDFSamplingRecord(const Intersection &its, const Vector &wi, const Vector &wo, ETransportMode mode)
  : its(its), sampler(NULL), wi(wi), wo(wo), mode(mode),
  typeMask(BSDF::EAll), component(-1), sampledType(0), sampledComponent(-1) {}

class NeuralBsdf : public BSDF {
public: 	

	typedef BlockedArray< Float > LatentCoeffs;

	NeuralBsdf(const Properties &props): BSDF(props) { 

		m_numlatentcoeffs = props.getFloat("latentsize", 32);
		m_numlayers = props.getFloat("hiddenlayers", 4) + 1; 
		m_layerlength = props.getFloat("neurons", 256);
		std::string wrapMode = props.getString("wrapMode", "repeat");
		m_wrapModeU = parseWrapMode(props.getString("wrapModeU", wrapMode));
		m_wrapModeV = parseWrapMode(props.getString("wrapModeV", wrapMode)); 
		m_texturescale = props.getFloat("btfscale", 1.0f);
		m_valuesarelinear = props.getBoolean("linearvalues", true);
		m_fake_silhouettes = props.getBoolean("silhouettes", false);

		m_latents.resize(m_numlatentcoeffs);
		m_fclayers.resize(m_numlayers);
		m_bias.resize(m_numlayers);
  
		m_folder = Thread::getThread()->getFileResolver()->resolve( props.getString("materialfolder") );
		Log(EInfo, "Loading textures from folder \"%s\"", m_folder.filename().string().c_str());
		if (!fs::is_directory(m_folder))
			Log(EError, "Folder \"%s\" could not be found!", m_folder.string().c_str()); 


		// READ LATENT MAPS
		for (int i = 0; i < m_numlatentcoeffs; i++) {
			std::string fnm = "map-"+boost::lexical_cast<std::string>(i)+".exr";
			fs::path filename = m_folder / fnm;	
			readInputEXR(filename, m_latents[i]);
		}

		// READ DECODER LAYERS
		for (int i = 0; i<m_numlayers; i++) {
			std::string fnm = "layer-"+boost::lexical_cast<std::string>(i +1)+".exr";
			fs::path filename = m_folder / fnm;
			readInputEXR(filename, m_fclayers[i]);

			std::string fnm2 = "bias-"+boost::lexical_cast<std::string>(i +1)+".exr";
			fs::path filename2 = m_folder / fnm2;
			readInputEXR(filename2, m_bias[i]);
		}

		// READ HEIGHT FIELD
		cout<<"checking for height field, using silhouettes: "<< m_fake_silhouettes <<endl;
		m_displacement = false;
		fs::path hf = Thread::getThread()->getFileResolver()->resolve( props.getString("heightfield",  "none") );
		
		if (std::strcmp(hf.filename().string().c_str(), "none") != 0) {
	 		m_displacement = true; 
			Log(EInfo, "Loading heightfield from file \"%s\"", hf.filename().string().c_str()); 
			ref<FileStream> fs = new FileStream( hf, FileStream::EReadOnly); 			

			Bitmap* bitmap = new Bitmap(Bitmap::EAuto, fs);  
			m_heightfield = new Float[ bitmap->getHeight() * bitmap->getWidth() ]; 
			m_zmin = 10000; m_zmax = -10000;

			for (int si=0; si<bitmap->getHeight(); si++) {
				for (int sj=0; sj<bitmap->getWidth(); sj++) { 			
					m_heightfield[si * bitmap->getWidth() + sj] = (Float) (bitmap->getPixel(Point2i(sj,si))[0]); 
					m_zmin = std::min( m_heightfield[si * bitmap->getWidth() + sj] , m_zmin);
					m_zmax = std::max( m_heightfield[si * bitmap->getWidth() + sj] , m_zmax);  
				}
			}  
			cout<<"Displacement surface read. Maximum: "<< m_zmax <<" and minimum "<<m_zmin<<endl;
 		} 

		// READ MEAN AND STANDARD DEVIATION MAPS 
 		fs::path filename1 = m_folder / "norm-1.exr"; 
    	ref<FileStream> fs1 = new FileStream( filename1, FileStream::EReadOnly); 
		Bitmap* bitmap = new Bitmap(Bitmap::EAuto, fs1);
		m_imWidth = bitmap->getSize().x;
		m_imHeight = bitmap->getSize().y;
		readInputEXR(filename1, m_means);

		fs::path filename2 = m_folder / "norm-2.exr"; 
		readInputEXR(filename2, m_stds);


		cout<<"Initializing LIBXSMM matrix multiplication library"<<endl;
		libxsmm_init();
		fstfclayer = libxsmm_mmfunction<Float>(1, m_layerlength, (m_numlatentcoeffs+4));
		midfclayer = libxsmm_mmfunction<Float>(1, m_layerlength, m_layerlength);
		endfclayer = libxsmm_mmfunction<Float>(1, 3, m_layerlength);  
	}   

	void readInputEXR( fs::path filename, Float* & arr ) const {
		ref<FileStream> fs = new FileStream( filename, FileStream::EReadOnly);  
		Bitmap* bitmap = new Bitmap(Bitmap::EAuto, fs); 	

		arr = new Float[ bitmap->getHeight() * bitmap->getWidth() ];

		Float mini = 1000; Float maxi = -1000;
		for (int si=0; si<bitmap->getHeight(); si++) {
			for (int sj=0; sj<bitmap->getWidth(); sj++) {
				mini = std::min( (bitmap->getPixel(Point2i(sj,si)) )[0], mini);
				maxi = std::max( (bitmap->getPixel(Point2i(sj,si)) )[0], maxi);

				arr[si * bitmap->getWidth() + sj] = (Float) (bitmap->getPixel(Point2i(sj,si))[0]); 
			}
		} 
		cout<<"Reading file "<<filename.string().c_str()<<" , array stats:: Mini="<<mini<<"  maxi="<<maxi<<" with size "<<bitmap->getHeight()<<" "<<bitmap->getWidth()<<endl;  
	}

	inline ReconstructionFilter::EBoundaryCondition parseWrapMode(const std::string &wrapMode) {
		if (wrapMode == "repeat")
			return ReconstructionFilter::ERepeat;
		else if (wrapMode == "clamp")
			return ReconstructionFilter::EClamp;
		else if (wrapMode == "mirror")
			return ReconstructionFilter::EMirror;
		else if (wrapMode == "zero" || wrapMode == "black")
			return ReconstructionFilter::EZero;
		else if (wrapMode == "one" || wrapMode == "white")
			return ReconstructionFilter::EOne;
		else
			Log(EError, "Unknown wrap mode '%s' -- must be "
				"'repeat', 'clamp', 'black', or 'white'!", wrapMode.c_str());
		return ReconstructionFilter::EZero; // make gcc happy
	}

	NeuralBsdf(Stream *stream, InstanceManager *manager) : BSDF(stream, manager) {

	 	m_folder = stream->readString();
		Log(EDebug, "Unserializing BTF textures \"%s\"", m_folder.filename().string().c_str()); 
		m_wrapModeU = (ReconstructionFilter::EBoundaryCondition) stream->readUInt();
		m_wrapModeV = (ReconstructionFilter::EBoundaryCondition) stream->readUInt(); 
		m_texturescale = stream->readFloat();

		configure();
	}

	void configure() { 
		m_components.clear();  
		m_usesRayDifferentials = false;
		m_components.push_back(EDiffuseReflection | EFrontSide | ESpatiallyVarying );
		BSDF::configure();
	}  


	inline bool displaceUV(Float & uvx, Float & uvy, const TVector3<Float> &wi, Float R) const {

		Float vx = - wi.x;	Float vy = - wi.y;	Float vz = - wi.z;
		Float uvz0 = m_zmax+ EPSILON; 
		Float uvx0 = uvx + (m_zmax+ EPSILON)/wi.z * wi.x;	
		Float uvy0 = uvy + (m_zmax+ EPSILON)/wi.z * wi.y; 

		if (m_fake_silhouettes){ // Treat this as if the height field's max height was at 0 elevation along the positive surface normal, so the silhouettes are inside the local convex hull  
			uvx0 = uvx;	
			uvy0 = uvy; 	
		}
		bool warp_heightfield = true; // configurations for the silhouette mode 
		bool tile_heightfield = true;

 		if ((std::abs(vx) <= EPSILON) && (std::abs(vy) <= EPSILON))  return true; 

		Float x0 = math::floorToInt(uvx0) + (vx < 0); 		Float y0 = math::floorToInt(uvy0) + (vy < 0);  
		Float dx = std::abs(uvx0 - x0); 					Float dy = std::abs(uvy0 - y0);

		int incX = 0; 	int incY = 0; 	bool intersection = false;	bool advX = false;  bool advY = false;
		Float tIts = 0;		Float zOff = 0.0;
 		TRay<TPoint3<Float>, TVector3<Float> > r; 		
 		r.o = TPoint3<Float>(uvx0, uvy0, uvz0); 	
 		r.d = -wi; 

		int x0i = int(x0);    
		int y0i = int(y0);

		int _x0 = int(x0i + 1.0* ((vx > 0) - (vx < 0))* incX );			
		int _y0 = int(y0i + 1.0* ((vy > 0) - (vy < 0))* incY ); 
		int _x1 = int(x0i + 1.0* ((vx > 0) - (vx < 0))* (incX +1) );		
		int _y1 = int(y0i + 1.0* ((vy > 0) - (vy < 0))* (incY +1) ); 

		tIts = 10000000000; 	Float _u, _v, _t;
		TPoint3<Float> p1(0.0,0.0,0.0), p2(0.0,0.0,0.0), p3(0.0,0.0,0.0), p4(0.0,0.0,0.0), pp1(0.0,0.0,0.0), pp2(0.0,0.0,0.0); 

		while ( (!intersection) && ( std::abs(zOff) < (m_zmax - m_zmin + EPSILON) ) ) {

			if (m_fake_silhouettes && tile_heightfield && ((_x0 < 0) || (_x0>=m_imWidth) || (_y0 < 0) || (_y0>=m_imWidth) || (_x1 < 0) || (_x1>=m_imWidth) || (_y1 < 0) || (_y1>=m_imWidth) ))
				break;

			Float deltaZ11 = 0.0; Float deltaZ12 = 0.0;
			Float deltaZ21 = 0.0; Float deltaZ22 = 0.0;
			Float deltaZ13 = 0.0; Float deltaZ31 = 0.0;

			Float scaled_R = R * m_imHeight * m_texturescale;
			Float perim_dist = 3.14159 / 2.0; // to preserve arc length

			if ((R > EPSILON) && warp_heightfield) { // warp the heightfield according to local curvature (if convex)
				deltaZ11 =  scaled_R * ( 1 - math::safe_sqrt( 1.0 - ((_x0-x0i)*(_x0-x0i) + (_y0-y0i)*(_y0-y0i)) / (perim_dist*perim_dist * scaled_R * scaled_R ) ) ); 
				deltaZ21 =  scaled_R * ( 1 - math::safe_sqrt( 1.0 - ((_x1-x0i)*(_x1-x0i) + (_y0-y0i)*(_y0-y0i)) / (perim_dist*perim_dist * scaled_R * scaled_R ) ) ); 
				deltaZ12 =  scaled_R * ( 1 - math::safe_sqrt( 1.0 - ((_x0-x0i)*(_x0-x0i) + (_y1-y0i)*(_y1-y0i)) / (perim_dist*perim_dist * scaled_R * scaled_R ) ) ); 
				deltaZ22 =  scaled_R * ( 1 - math::safe_sqrt( 1.0 - ((_x1-x0i)*(_x1-x0i) + (_y1-y0i)*(_y1-y0i)) / (perim_dist*perim_dist * scaled_R * scaled_R ) ) );  
			}

			p1 = TPoint3<Float>(_x0, _y0, m_heightfield[ math::modulo(_x0, m_imHeight) * m_imHeight + math::modulo(_y0, m_imHeight) ]  - deltaZ11 ); 
			p2 = TPoint3<Float>(_x0, _y1, m_heightfield[ math::modulo(_x0, m_imHeight) * m_imHeight + math::modulo(_y1, m_imHeight) ]  - deltaZ12 ); 
			p3 = TPoint3<Float>(_x1, _y0, m_heightfield[ math::modulo(_x1, m_imHeight) * m_imHeight + math::modulo(_y0, m_imHeight) ]  - deltaZ21 ); 
			p4 = TPoint3<Float>(_x1, _y1, m_heightfield[ math::modulo(_x1, m_imHeight) * m_imHeight + math::modulo(_y1, m_imHeight) ]  - deltaZ22 );

			if (Triangle::rayIntersect( p3, p4, p2, r, _u, _v, _t )) tIts = std::min(tIts, _t);
			if (Triangle::rayIntersect( p1, p3, p2, r, _u, _v, _t )) tIts = std::min(tIts, _t);   
			if (tIts < 10000000) intersection = true; 
	 
 			// check which direction gets incremented to check the next square 
			if ( std::abs( vy * (incX+1 - dx) ) < std::abs( vx * (incY+1 - dy) ) ) {	
				advX = true;
				zOff = vz * std::abs(1.0*incX / vx);
			} 
			if ( std::abs( vy * (incX+1 - dx) ) > std::abs( vx * (incY+1 - dy) ) ) {
				advY = true; 
				zOff = vz * std::abs(1.0*incY / vy);
			}
			if ( std::abs( vy * (incX+1 - dx) ) == std::abs( vx * (incY+1 - dy) ) ) {
				
				int _x2 = int( x0i + 1.0* ((vx > 0) - (vx < 0))* (incX+2) );		
				int _y2 = int( y0i + 1.0* ((vy > 0) - (vy < 0))* (incY+2) );  

				if ((R > EPSILON) && warp_heightfield) { 
					deltaZ13 =  scaled_R * ( 1 - math::safe_sqrt( 1.0 - ((_x1-x0i)*(_x1-x0i) + (_y2-y0i)*(_y2-y0i)) / (perim_dist * perim_dist * scaled_R *scaled_R) ) ); 
					deltaZ31 =  scaled_R * ( 1 - math::safe_sqrt( 1.0 - ((_x2-x0i)*(_x2-x0i) + (_y1-y0i)*(_y1-y0i)) / (perim_dist * perim_dist * scaled_R *scaled_R) ) );  
				}

				pp1 = TPoint3<Float>(_x2, _y1, m_heightfield[ math::modulo(_x2, m_imHeight) * m_imHeight + math::modulo(_y1, m_imHeight) ]  - deltaZ31  ); 
				pp2 = TPoint3<Float>(_x1, _y2, m_heightfield[ math::modulo(_x1, m_imHeight) * m_imHeight + math::modulo(_y2, m_imHeight) ]  - deltaZ13  ); 

				if (Triangle::rayIntersect( p2, p4, pp2, r, _u, _v, _t )) tIts = std::min(tIts, _t);
				if (Triangle::rayIntersect( p4, p3, pp1, r, _u, _v, _t )) tIts = std::min(tIts, _t);   
				if (tIts < 10000000) intersection = true;  

				advX = true;  advY = true;
				zOff = vz * std::abs(1.0*incX / vx); 
			} 
			if (advX) incX++;
			if (advY) incY++;  

			advX = false; 	advY = false;
			_x0 = int(x0i + 1.0 *((vx > 0) - (vx < 0))* incX );			
			_y0 = int(y0i + 1.0 *((vy > 0) - (vy < 0))* incY ); 
			_x1 = int(x0i + 1.0 *((vx > 0) - (vx < 0))* (incX +1) );		
			_y1 = int(y0i + 1.0 *((vy > 0) - (vy < 0))* (incY +1) ); 

			if ( (R > EPSILON) && ( sqrt( (_x0-x0i)*(_x0-x0i) + (_y0-y0i)*(_y0-y0i) ) > perim_dist * scaled_R ) )  // iterated further than the radius of the sphere tangent to the mesh
				break; // trick to generate silhouettes
		} 
		if (intersection) { 
			uvx = math::modulo(uvx0 + vx * tIts, Float(m_imHeight));
			uvy = math::modulo(uvy0 + vy * tIts, Float(m_imHeight));   
		} 
		return intersection; 
	} 


	Spectrum eval(const BSDFSamplingRecord &bRec, EMeasure measure) const {  

		if ((Frame::cosTheta(bRec.wi) < EPSILON) && (m_fake_silhouettes) && (std::abs(1-dot(bRec.wi, -bRec.wo)) < EPSILON)) {  
			return Spectrum(1.0f);  
		} 

		if( ( Frame::cosTheta(bRec.wi) >= EPSILON ) && ( Frame::cosTheta(bRec.wo) >= EPSILON) ) { 

			Float u = bRec.its.uv.x;
			Float v = bRec.its.uv.y; 
			TVector3<Float> wi = bRec.wi; 
			TVector3<Float> wo = bRec.wo;			

			if (bRec.sampledComponent != 0) { // means the eval method is called directly, rather than from inside the sample method 
				u = math::modulo( math::modulo(bRec.its.uv.x, Float(1.0)) * m_imWidth * m_texturescale, Float(m_imWidth) ); 
				v = math::modulo( math::modulo(bRec.its.uv.y, Float(1.0)) * m_imHeight * m_texturescale, Float(m_imHeight) ); 

				if (math::signum(wi.x) != math::signum( dot(bRec.its.shFrame.toWorld(bRec.wi), bRec.its.dpdu )) )
					wi.x = -wi.x; 
				if (math::signum(wi.y) != math::signum( dot(bRec.its.shFrame.toWorld(bRec.wi), bRec.its.dpdv )) )
					wi.y = -wi.y;   

				if (m_displacement) {  
					Float R = 0.0;
					if (m_fake_silhouettes) {

						// Formulas taken from https://mathworld.wolfram.com/FirstFundamentalForm.html
						// https://mathworld.wolfram.com/SecondFundamentalForm.html
						// https://mathworld.wolfram.com/NormalCurvature.html

						TVector3<Float> dndu, dndv;
						bRec.its.shape->getNormalDerivative(bRec.its, dndu, dndv, true);
						
						Float a = wi.x / sqrt(wi.x*wi.x + wi.y*wi.y);
						Float b = wi.y / sqrt(wi.x*wi.x + wi.y*wi.y); 

						Float E =  dot(bRec.its.dpdu, bRec.its.dpdu);
						Float F =  dot(bRec.its.dpdu, bRec.its.dpdv);
						Float G =  dot(bRec.its.dpdv, bRec.its.dpdv);
						Float e = -dot(bRec.its.dpdu, dndu);
						Float f = -dot(bRec.its.dpdu, dndv);
						Float g = -dot(bRec.its.dpdv, dndv);
						Float K = ( e*a*a + 2*f*a*b + g*b*b ) / ( E*a*a + 2*F*a*b + G*b*b );
						
						K = -K; 
						if ( K > EPSILON) 
							R = 1.0 / K;  // radius of curvature  

						Float uvnorm2 = dot( a* bRec.its.dpdu + b* bRec.its.dpdv, a* bRec.its.dpdu + b* bRec.its.dpdv );
						R = R / sqrt(uvnorm2);
					}

					bool hit = displaceUV(u, v, wi,  R);   

					if (!hit){ 
						return Spectrum( 0.0f); 
					} 
				} 
			}  

			int xPos = math::floorToInt(u); 
			int yPos = math::floorToInt(v);
			Float dx1 = u - xPos, dx2 = 1.0f - dx1;  
			Float dy1 = v - yPos, dy2 = 1.0f - dy1; 

			std::vector<Float> lats;
			for (int i=0; i<m_numlatentcoeffs; i++) {

				Float val = m_latents[i][ math::modulo(xPos, m_imWidth) * m_imHeight + math::modulo(yPos, m_imHeight) ] * dx2 * dy2 +
					m_latents[i][ math::modulo(xPos, m_imWidth) * m_imHeight + math::modulo(yPos+1, m_imHeight) ] * dx2 * dy1 +
					m_latents[i][ math::modulo(xPos+1, m_imWidth) * m_imHeight + math::modulo(yPos, m_imHeight) ] * dx1 * dy2 +
					m_latents[i][ math::modulo(xPos+1, m_imWidth) * m_imHeight + math::modulo(yPos+1, m_imHeight) ] * dx1 * dy1; 
				lats.push_back( val );  
			}
 
			Float r = 0.0f;
			Float g = 0.0f;
			Float b = 0.0f;
			computeNN(wi, wo, lats, r, g, b);  

			Float mean = m_means[ math::modulo(xPos, m_imWidth) * m_imHeight + math::modulo(yPos, m_imHeight) ] * dx2 * dy2 +
					m_means[ math::modulo(xPos, m_imWidth) * m_imHeight + math::modulo(yPos+1, m_imHeight) ] * dx2 * dy1 +
					m_means[ math::modulo(xPos+1, m_imWidth) * m_imHeight + math::modulo(yPos, m_imHeight) ] * dx1 * dy2 +
					m_means[ math::modulo(xPos+1, m_imWidth) * m_imHeight + math::modulo(yPos+1, m_imHeight) ] * dx1 * dy1; 

			Float std = m_stds[ math::modulo(xPos, m_imWidth) * m_imHeight + math::modulo(yPos, m_imHeight) ] * dx2 * dy2 +
					m_stds[ math::modulo(xPos, m_imWidth) * m_imHeight + math::modulo(yPos+1, m_imHeight) ] * dx2 * dy1 +
					m_stds[ math::modulo(xPos+1, m_imWidth) * m_imHeight + math::modulo(yPos, m_imHeight) ] * dx1 * dy2 +
					m_stds[ math::modulo(xPos+1, m_imWidth) * m_imHeight + math::modulo(yPos+1, m_imHeight) ] * dx1 * dy1; 

			r = std::exp( r * std + mean) -1.0 ;
			g = std::exp( g * std + mean) -1.0 ;
			b = std::exp( b * std + mean) -1.0 ;

			if (!m_valuesarelinear) {
				r = std::pow(r, 2.2);
				g = std::pow(g, 2.2);
				b = std::pow(b, 2.2);
			}

			Spectrum res;  
			res.fromLinearRGB( std::max(r, 0.0f), std::max(g, 0.0f), std::max(b, 0.0f) );
			return res  * std::abs(Frame::cosTheta(bRec.wo));    	  
		} 
		return Spectrum(0.0f);  
	} 

	
	void computeNN(const TVector3<Float> & wi, const TVector3<Float> & wo, std::vector<Float> & lats, Float& r, Float& g, Float& b) const {
		float vu = wi.x / (1 + wi.z);
		float vv = wi.y / (1 + wi.z);
		float lu = wo.x / (1 + wo.z);
		float lv = wo.y / (1 + wo.z);
		lats.push_back(vu);
		lats.push_back(vv);
		lats.push_back(lu);
		lats.push_back(lv);
		Float curr[m_layerlength] = {0}; 
		Float prev[m_layerlength] = {0}; 

		std::copy(&lats[0], &lats[0]+m_numlatentcoeffs+4, prev); 
		std::copy(m_bias[0], m_bias[0]+m_layerlength, curr);  
		fstfclayer(prev, m_fclayers[0], curr);

		for (int i=0; i<m_layerlength; i++) 
			prev[i] = std::max(curr[i], Float(0.0f));  

		for (int ll=1; ll<m_numlayers-1; ll++) { 
			std::copy(m_bias[ll], m_bias[ll]+m_layerlength, curr); 
			midfclayer(prev, m_fclayers[ll], curr);

			for (int i=0; i<m_layerlength; i++) 
				prev[i] = std::max(curr[i], Float(0.0f)); 
		} 
		std::copy(m_bias[m_numlayers-1], m_bias[m_numlayers-1]+3, curr);
		endfclayer(prev, m_fclayers[m_numlayers-1], curr);

		r = curr[0];
		g = curr[1];
		b = curr[2];
	}  


	Float pdf(const BSDFSamplingRecord &bRec, EMeasure measure) const {    
		if ( Frame::cosTheta(bRec.wi) <= 0 || Frame::cosTheta(bRec.wo) <= 0 || measure != ESolidAngle)
			return 0.0f;
		return warp::squareToCosineHemispherePdf(bRec.wo);
	}

 
	Spectrum sample(BSDFSamplingRecord &bRec, Float &_pdf, const Point2 &sample) const {  

		// Transparent backside for silhouettes
		if ((Frame::cosTheta(bRec.wi) <= EPSILON) && (m_fake_silhouettes)) { 
			_pdf = 1.0f; 
			bRec.wo = - bRec.wi; 
			return Spectrum(1.0f);  
		} 

		bRec.wo = warp::squareToCosineHemisphere(sample);   

		Float u = math::modulo( math::modulo(bRec.its.uv.x, Float(1.0)) * m_imWidth * m_texturescale, Float(m_imWidth) ); 
		Float v = math::modulo( math::modulo(bRec.its.uv.y, Float(1.0)) * m_imHeight * m_texturescale, Float(m_imHeight) ); 

		Intersection newits; 
		TVector3<Float> wi = bRec.wi; 
		TVector3<Float> wo = bRec.wo;

		if (math::signum(wi.x) != math::signum( dot(bRec.its.shFrame.toWorld(bRec.wi), bRec.its.dpdu)) )
			wi.x = -wi.x; 
		if (math::signum(wi.y) != math::signum( dot(bRec.its.shFrame.toWorld(bRec.wi), bRec.its.dpdv)) )
			wi.y = -wi.y;   

		if (m_displacement) {  
			Float R = 0.0;

			if (m_fake_silhouettes) {
				// Formulas taken from https://mathworld.wolfram.com/FirstFundamentalForm.html
				// https://mathworld.wolfram.com/SecondFundamentalForm.html 
				// https://mathworld.wolfram.com/NormalCurvature.html

				TVector3<Float> dndu, dndv;
				bRec.its.shape->getNormalDerivative(bRec.its, dndu, dndv, true);
				Float a = wi.x / sqrt(wi.x*wi.x + wi.y*wi.y);
				Float b = wi.y / sqrt(wi.x*wi.x + wi.y*wi.y);

				Float E =  dot(bRec.its.dpdu, bRec.its.dpdu);
				Float F =  dot(bRec.its.dpdu, bRec.its.dpdv);
				Float G =  dot(bRec.its.dpdv, bRec.its.dpdv);
				Float e = -dot(bRec.its.dpdu, dndu);
				Float f = -dot(bRec.its.dpdu, dndv);
				Float g = -dot(bRec.its.dpdv, dndv);

				Float K = ( e*a*a + 2*f*a*b + g*b*b ) / ( E*a*a + 2*F*a*b + G*b*b );
				K = -K; 
				if ( K > EPSILON) 
					R = 1.0 / K;  // radius of curvature 

				Float uvnorm2 = dot( a* bRec.its.dpdu + b* bRec.its.dpdv, a* bRec.its.dpdu + b* bRec.its.dpdv );
				R = R / sqrt(uvnorm2);
			} 

			bool hit = displaceUV(u, v, wi,  R);   

			if (!hit){ 
				_pdf = 1.0f; 
				bRec.wo = - bRec.wi; 
				return Spectrum( 1.0f);  
			} 
		} 

		newits.uv = Point2(u,v);  
		BSDFSamplingRecord bRec_mod(newits, wi, wo, bRec.mode);
		bRec_mod.sampledComponent = 0;

		_pdf = pdf(bRec_mod, ESolidAngle); 
		if (_pdf == 0) 
			return Spectrum(0.0f); 

		return eval(bRec_mod, ESolidAngle) / _pdf; 
	}

 
	Spectrum sample(BSDFSamplingRecord &bRec, const Point2 &sample) const {
		Float pdf;
		return NeuralBsdf::sample(bRec, pdf, sample);
	} 

	void addChild(const std::string &name, ConfigurableObject *child) {}

	void serialize(Stream *stream, InstanceManager *manager) const {
		BSDF::serialize(stream, manager); 
		stream->writeString(m_folder.string());
		stream->writeUInt(m_wrapModeU);
		stream->writeUInt(m_wrapModeV); 
		stream->writeFloat(m_texturescale);  
	}  

	std::string toString() const {
		std::ostringstream oss;
		oss << "NeuralBsdf[" << endl
   			<< "  id = \"" << getID() << "\"," << endl
			<< "  btf folder = " << m_folder.filename().string().c_str() << endl 
			<< "]";
		return oss.str();
	}

	Shader *createShader(Renderer *renderer) const;

	MTS_DECLARE_CLASS()
private:

	libxsmm_mmfunction<Float> fstfclayer;
	libxsmm_mmfunction<Float> midfclayer;
	libxsmm_mmfunction<Float> endfclayer; 

	ReconstructionFilter::EBoundaryCondition m_wrapModeU;
	ReconstructionFilter::EBoundaryCondition m_wrapModeV; 

	Float m_texturescale;
	fs::path m_folder;  
	std::vector<Float *> m_latents;
	std::vector<Float *> m_fclayers;
	std::vector<Float *> m_bias;
	Float* m_means;
	Float* m_stds;

	int m_numlatentcoeffs;
	int m_numlayers;
	int m_layerlength;

	bool m_displacement;
	Float * m_heightfield;

	bool m_fake_silhouettes;

	Float m_zmin;
	Float m_zmax;

	int m_imWidth;
	int m_imHeight;
	bool m_valuesarelinear;

	const float EPSILON = 0.0000001;
};

// ================ Hardware shader implementation ================  THIS IS NOT USED AT THE MOMENT
class NeuralBsdfShader : public Shader {
public:
	NeuralBsdfShader(Renderer *renderer)
		  : Shader(renderer, EBSDFShader) {}

	void putDependencies(std::vector<Shader *> &deps) {}

	void cleanup(Renderer *renderer) {}

	void generateCode(std::ostringstream &oss,
			const std::string &evalName,
			const std::vector<std::string> &depNames) const {}

	MTS_DECLARE_CLASS()
private:  
};

Shader *NeuralBsdf::createShader(Renderer *renderer) const { 
	return new NeuralBsdfShader(renderer);
}

MTS_IMPLEMENT_CLASS(NeuralBsdfShader, false, Shader)
MTS_IMPLEMENT_CLASS_S(NeuralBsdf, false, BSDF)
MTS_EXPORT_PLUGIN(NeuralBsdf, "BSDF for Neural Network BTF shading");
MTS_NAMESPACE_END