import numpy as np
import torch
import torch.nn as nn
from torch.autograd import Variable
import sys
import array
import OpenEXR
import Imath
import glob
import os 
from functools import reduce
from random import shuffle  
from collections import OrderedDict
from math import exp


def write_exr_gray(im, pth):
    im32 = np.float32(im)
    y = im32[:,:].tostring()  
    print(im32.shape, im32.dtype) 
    header = OpenEXR.Header(im32.shape[0], im32.shape[1]) 
    header['channels'] = { 'Y' : Imath.Channel(Imath.PixelType(OpenEXR.FLOAT))} 
    out = OpenEXR.OutputFile(pth, header) 
    out.writePixels({'Y' : y})

def to_var(x):
    if torch.cuda.is_available():
        x = x.cuda()
    return Variable(x)  

def load_list_exr(pth):
    imexr = OpenEXR.InputFile(pth) 
    (r, g, b, vaz, vel, laz, lel) = imexr.channels("RGBCDEF")
    dw = imexr.header()['dataWindow']
    imsize = (dw.max.x - dw.min.x + 1, dw.max.y - dw.min.y + 1)
    im = np.zeros( (imsize[0],imsize[1], 7), dtype=np.float32)     
    im[:,:,0] = np.reshape( np.frombuffer(r, dtype=np.float32), imsize)
    im[:,:,1] = np.reshape( np.frombuffer(g, dtype=np.float32), imsize)
    im[:,:,2] = np.reshape( np.frombuffer(b, dtype=np.float32), imsize)
    im[:,:,3] = np.reshape( np.frombuffer(vaz, dtype=np.float32), imsize)
    im[:,:,4] = np.reshape( np.frombuffer(vel, dtype=np.float32), imsize)
    im[:,:,5] = np.reshape( np.frombuffer(laz, dtype=np.float32), imsize)
    im[:,:,6] = np.reshape( np.frombuffer(lel, dtype=np.float32), imsize) 
    return im 


def import_lists( listinputpaths, outputpath, maxnumofims , shuffle = False, logtransform = True, sparsity=False, separatetraintest=False, whitening=True, verbose=True):
# reads abrdf texels, 
# does stereographic projection of angles (from spherical angles in degrees to stereographic u-v), 
# applies log transform and standardizes the input data
# splits data into training and testing once it's imported

# this code needs cleaning, it has been simplified from a previous version that handled input with different angular samplings

    if not os.path.exists(outputpath):
        os.makedirs(outputpath)   
    
    if (verbose):
        print(listinputpaths)
    
    fnamesl = (glob.glob(listinputpaths+"/*.exr"))
    fnamesl.sort(key=lambda x: x.split("/")[-1], reverse=False)
    pxind = 0 
    prevind = 0
    templist = []
    flist = []
    i = 0

    while((i < len(fnamesl)) and (i<maxnumofims)):
        ftem = fnamesl[i].split(".exr") 
        finfo = ftem[0].split("-")  
        pxind = int(finfo[2]) 
        flist.append(fnamesl[i]) 
        i = i+1  
    if (verbose):
        print('number of pxs: ', len(flist))
    
    listsdataset = []
    listangles = [] 
    
    im0 = load_list_exr(flist[0])   
    s = np.array(im0.shape) 
        
    norms = np.zeros( (min(maxnumofims, len(flist)), 2), dtype=np.float32)  
    listsdataset = np.zeros( (min( maxnumofims, len(flist)), s[0], 1, 3), dtype=np.float32)  
    listangles = np.zeros( (min( maxnumofims, len(flist)), s[0], 1, 4), dtype=np.float32)  
    
    pxids = np.arange(len(flist))    
    if shuffle:
        np.random.shuffle(pxids) 
    datamin = 100000.0
    datamax = -1000.0
    
    for i in range(min(maxnumofims, len(flist))):
         
        vlen = 0 
        imo = load_list_exr(flist[pxids[i]])  
        im = imo[:,:,0:3]
        
        ima = np.zeros( (imo.shape[0], imo.shape[1], 4), dtype=np.float32)
        vx = np.cos(np.pi /180.0 * imo[:,:,3]) * np.cos(np.pi /180.0 * imo[:,:,4])
        vy = np.sin(np.pi /180.0 * imo[:,:,3]) * np.cos(np.pi /180.0 * imo[:,:,4])
        vz = np.sin(np.pi /180.0 * imo[:,:,4])
        vu = vx / (1 + vz) 
        vv = vy / (1 + vz) 
        vx2 = np.cos(np.pi /180.0 * imo[:,:,5]) * np.cos(np.pi /180.0 * imo[:,:,6])
        vy2 = np.sin(np.pi /180.0 * imo[:,:,5]) * np.cos(np.pi /180.0 * imo[:,:,6])
        vz2 = np.sin(np.pi /180.0 * imo[:,:,6])
        vu2 = vx2 / (1 + vz2) 
        vv2 = vy2 / (1 + vz2) 
        ima[:,:,0] = vu
        ima[:,:,1] = vv
        ima[:,:,2] = vu2
        ima[:,:,3] = vv2
        
        datamin = min(datamin, np.min(im))
        datamax = max(datamax, np.max(im)) 
        
        if logtransform:
            im = np.log( np.maximum(im, 0.0) + 1.0 )
        
        norms[i, 0] = np.mean(im)
        norms[i, 1] = np.std(im) 
        if whitening:
            im = (im - norms[i,0]) / norms[i,1] 
        
        listsdataset[i, :,:,:] = im
        listangles[i, :,:,:] = ima 
        
        if (verbose):
            if ((i+1)%250==0):
                print('read in ', (i+1), ' pxlist images')
            
    if (verbose):
        print('read in datasets')
    listsdataset = torch.from_numpy( listsdataset ).squeeze_()
    listangles = torch.from_numpy( listangles ).squeeze_()
    if (verbose):
        print( listsdataset.size(), listangles.size(), ' minmax: ', datamin, datamax )   
    
    if separatetraintest:
        sliceidx = int(0.8 * maxnumofims) 
        trainlists = [] 
        testlists = [] 
        trainangles = []
        testangles = [] 

        trainlists = listsdataset[ 0:sliceidx] 
        trainangles = listangles[ 0:sliceidx]
        testlists = listsdataset[ sliceidx: len(listsdataset)]
        testangles = listangles[ sliceidx: len(listangles)]

        if (verbose):
            print(len(trainlists), len(trainangles), len(testlists), len(testangles))

        return trainlists, trainangles, testlists, testangles, norms, pxids
    
    return listsdataset, listangles, norms, pxids



class unifiednet(nn.Module):
    def __init__(self, mlplayersin, encoderlayersin, decoderlayersin, kerneldim, latentsize): 
        super(unifiednet, self).__init__() 
        self.latentsize = latentsize
        self.kerneldim = kerneldim 

        self.activate = nn.PReLU(3 * self.kerneldim)      
        self.wbias = nn.Parameter( (torch.zeros([self.kerneldim], dtype=torch.float32) ).unsqueeze_(0).unsqueeze_(0) ) 
        
        mlplayers = OrderedDict() 
        ii = 0
        mlplayers[str(0)] = nn.Linear(4, mlplayersin[0])
        mlplayers[str(1)] = nn.ReLU(True)
        ii = 2                
        for i in range(len(mlplayersin)-1):
            mlplayers[str(ii)] = nn.Linear(mlplayersin[i],mlplayersin[i+1])
            ii = ii + 1
            mlplayers[str(ii)] = nn.ReLU(True)
            ii = ii + 1 
            
        mlplayers[str(ii)] = nn.Linear(mlplayersin[-1], 3 * self.kerneldim)
        self.mlp = nn.Sequential(mlplayers) 
        
        enclayers = OrderedDict() 
        ii = 0
        if len(encoderlayersin) >0:      
            enclayers[str(0)] = nn.Linear(3*self.kerneldim, encoderlayersin[0])
            enclayers[str(1)] = nn.ReLU(True)
            ii = 2       
            for i in range(len(encoderlayersin)-1):
                enclayers[str(ii)] = nn.Linear(encoderlayersin[i],encoderlayersin[i+1])
                ii = ii + 1
                enclayers[str(ii)] = nn.ReLU(True)
                ii = ii + 1

            enclayers[str(ii)] = nn.Linear(encoderlayersin[-1], self.latentsize)        
        else:
            enclayers[str(0)] = nn.Linear(3*self.kerneldim, self.latentsize)
        self.encoder = nn.Sequential(enclayers) 
        
        declayers = OrderedDict() 
        declayers[str(0)] = nn.Linear(latentsize + 4, decoderlayersin[0])
        declayers[str(1)] = nn.ReLU(True)
        ii = 2      
        for i in range(len(decoderlayersin)-1):
            declayers[str(ii)] = nn.Linear(decoderlayersin[i],decoderlayersin[i+1])
            ii = ii + 1
            declayers[str(ii)] = nn.ReLU(True) 
            ii = ii + 1             
        declayers[str(ii)] = nn.Linear(decoderlayersin[-1], 3)
        self.generator = nn.Sequential(declayers) 
        
        print("Number of params in the encoder ", sum(p.numel() for p in self.encoder.parameters() if p.requires_grad))  
        print("Number of params in the decoder ", sum(p.numel() for p in self.generator.parameters() if p.requires_grad))  
        for m in self.modules(): 
            if isinstance(m, nn.Conv2d): 
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels 
                m.weight.data.normal_(0, np.sqrt(2. / n)) 
            elif isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.BatchNorm1d): 
                m.weight.data.fill_(1) 
                m.bias.data.zero_()    
            if isinstance(m, nn.Linear): 
                nn.init.xavier_uniform(m.weight.data) 
                
                
    def forward(self, xin):  
        w = self.mlp(xin[1][0,0:xin[0].size()[1],:]).unsqueeze_(0).unsqueeze_(3).view(1, xin[0].size()[1], 3, self.kerneldim)
        w = w.expand(xin[0].size()[0], -1, -1, -1)         
        xlat = self.encoder( self.activate((torch.mean( xin[0].unsqueeze_(3).expand(-1, -1, -1, self.kerneldim) * w, 1 ) + self.wbias.expand(xin[0].size()[0], 3, -1) ).view(xin[0].size()[0], 3*self.kerneldim) ) )
        return self.generator(torch.cat( (xlat.unsqueeze_(1).expand(-1, xin[1].size()[1], -1), xin[1] ), dim = 2 ).view( -1, self.latentsize + 4 ) ).view( xin[0].size()[0], xin[1].size()[1], 3 )
    
    
    def encode(self, xin): # only works for a single input
        w = self.mlp(xin[1][0,0:xin[0].size()[1],:]).unsqueeze_(0).unsqueeze_(3).view(1, xin[0].size()[1], 3, self.kerneldim)
        w2 = w.expand(xin[0].size()[0], -1, -1, -1)
        x1 = xin[0].unsqueeze_(3).expand(-1, -1, -1, self.kerneldim)  
        x2 = self.activate( (torch.mean( x1* w2, 1 ) + self.wbias.expand(xin[0].size()[0], 3, -1)).view(xin[0].size()[0], 3*self.kerneldim) )  
        return self.encoder(x2).squeeze_()
